;!function(){try{var e="undefined"!=typeof window?window:"undefined"!=typeof global?global:"undefined"!=typeof self?self:{},n=(new Error).stack;n&&(e._btDebugIds=e._btDebugIds||{},e._btDebugIds[n]="f3275c4b-b90b-2366-23eb-b791c77c6d69")}catch(e){}}();
import { AlbumType } from '../photos-library/model/album.js';
import * as fs from 'fs/promises';
import * as path from 'path';
import { iCPSError } from '../../app/error/error.js';
import { ARCHIVE_ERR } from '../../app/error/error-codes.js';
import { Resources } from '../resources/main.js';
import { iCPSEventArchiveEngine, iCPSEventRuntimeWarning } from '../resources/events-types.js';
/**
 * This class handles the archiving of albums
 */
export class ArchiveEngine {
    /**
     * The local photos library
     */
    photosLibrary;
    /**
     * The remote iCloud connection
     */
    icloud;
    /**
     * Creates a new Archive Engine object
     * @param icloud - The iCloud connection
     * @param photosLibrary - The local photos library
     */
    constructor(icloud, photosLibrary) {
        this.icloud = icloud;
        this.photosLibrary = photosLibrary;
    }
    /**
     * This will archive an Album stored in the given location and delete it's remote representation, if remoteDelete is set
     * @remarks This function expects the asset to be located in the primary zone - shared assets in folders is currently not supported by the api
     * @param archivePath - The path to the local album. The named path is expected.
     * @param assetList - The current remote asset list
     * @returns A Promise, that resolves once the path has been archived
     * @throws An iCPSError if the album could not be archived
     * @emits ARCHIVE_START - When the archiving starts
     * @emits PERSISTING_START - When the persisting starts
     * @emits REMOTE_DELETE - When the remote deletion starts
     * @emits ARCHIVE_DONE - When the archiving is done
     * @emits ARCHIVE_ASSET_ERROR - When an error occurs while persisting an asset
     */
    async archivePath(archivePath, assetList) {
        Resources.logger(this).debug(`Archiving path ${archivePath}`);
        Resources.emit(iCPSEventArchiveEngine.ARCHIVE_START, archivePath);
        if (assetList.length === 0) {
            throw new iCPSError(ARCHIVE_ERR.NO_ASSETS);
        }
        const albumName = path.basename(archivePath);
        if (albumName.startsWith(`.`)) {
            throw new iCPSError(ARCHIVE_ERR.UUID_PATH);
        }
        const parentFolderPath = path.dirname(archivePath);
        const [archivedAlbum, archivedAlbumPath] = await this.photosLibrary.readFolderFromDisk(albumName, parentFolderPath, ``);
        if (archivedAlbum.albumType !== AlbumType.ALBUM) {
            throw new iCPSError(ARCHIVE_ERR.NON_ALBUM);
        }
        const loadedAlbum = (await this.photosLibrary.loadAlbum(archivedAlbum, archivedAlbumPath)).find(album => album.albumName === albumName);
        if (!loadedAlbum || Object.keys(loadedAlbum.assets).length === 0) {
            throw new iCPSError(ARCHIVE_ERR.LOAD_FAILED);
        }
        const numberOfItems = Object.keys(loadedAlbum.assets).length;
        Resources.logger(this).debug(`Persisting ${numberOfItems} items`);
        Resources.emit(iCPSEventArchiveEngine.PERSISTING_START, numberOfItems);
        // Iterating over all album items to persist them
        const remoteDeleteList = await Promise.all(Object.keys(loadedAlbum.assets).map(async (uuidFilename) => {
            const assetPath = path.join(this.photosLibrary.primaryAssetDir, uuidFilename);
            const archivedAssetPath = path.join(archivedAlbumPath, loadedAlbum.assets[uuidFilename]);
            try {
                await this.persistAsset(assetPath, archivedAssetPath);
                return this.prepareForRemoteDeletion(assetPath, assetList);
            }
            catch (err) {
                Resources.emit(iCPSEventRuntimeWarning.ARCHIVE_ASSET_ERROR, new iCPSError(ARCHIVE_ERR.PERSIST_FAILED)
                    .addCause(err)
                    .addContext(`assetPath`, assetPath)
                    .addContext(`archivedAssetPath`, archivedAssetPath), assetPath);
                return undefined;
            }
        }));
        // Filtering for unique & undefined entries
        const uniqueDeleteList = [...new Set(remoteDeleteList.filter(obj => obj !== undefined))];
        if (uniqueDeleteList && uniqueDeleteList.length > 0 && Resources.manager().remoteDelete) {
            try {
                Resources.emit(iCPSEventArchiveEngine.REMOTE_DELETE, uniqueDeleteList.length);
                await this.icloud.photos.deleteAssets(uniqueDeleteList);
            }
            catch (err) {
                throw new iCPSError(ARCHIVE_ERR.REMOTE_DELETE_FAILED)
                    .addCause(err);
            }
        }
        Resources.emit(iCPSEventArchiveEngine.ARCHIVE_DONE);
    }
    /**
     * Persists a locally cached asset in the proper folder - applying original files m & a times accordingly
     * @param assetPath - Path to the assets file path in the Assets folder
     * @param archivedAssetPath - The target path of the asset (with filename)
     * @returns A Promise that resolves, once the file has been copied or rejects, in case there was an error
     */
    async persistAsset(assetPath, archivedAssetPath) {
        Resources.logger(this).debug(`Persisting ${assetPath} to ${archivedAssetPath}`);
        const fileStat = await fs.stat(assetPath);
        // Const lFileStat = await fs.lstat(archivedAssetPath)
        await fs.unlink(archivedAssetPath);
        await fs.copyFile(assetPath, archivedAssetPath);
        await fs.utimes(archivedAssetPath, fileStat.mtime, fileStat.mtime);
    }
    /**
     * Prepares the asset for deletion
     * @param assetPath - The path to the assets location in the Assets folder
     * @param assetList - A full list of all remote assets
     * @returns A string containing the asset's recordName or undefined, if the asset should not be deleted
     * @throws An iCPSError if loading fails
     */
    prepareForRemoteDeletion(assetPath, assetList) {
        if (!Resources.manager().remoteDelete) {
            return undefined;
        }
        // Get asset UUID to find record name from assetList
        const assetUUID = Buffer.from(path.parse(assetPath).name, `base64url`).toString(`base64`);
        const asset = assetList.find(asset => asset.getUUID() === assetUUID);
        if (!asset) {
            throw new iCPSError(ARCHIVE_ERR.NO_REMOTE_ASSET)
                .addMessage(assetUUID)
                .addContext(`assetList`, assetList);
        }
        if (!asset.recordName) {
            throw new iCPSError(ARCHIVE_ERR.NO_REMOTE_RECORD_NAME)
                .addMessage(asset.getDisplayName())
                .addContext(`asset`, asset);
        }
        if (asset.isFavorite) {
            Resources.logger(this).debug(`Not deleting favorite asset ${asset.getDisplayName()}`);
            return undefined;
        }
        Resources.logger(this).debug(`Returning asset ${asset.recordName} for deletion`);
        return asset.recordName;
    }
}
//# sourceMappingURL=archive-engine.js.map
//# debugId=f3275c4b-b90b-2366-23eb-b791c77c6d69